local math_min = math.min
local math_max = math.max
Hooks:PostHook(PlayerManager,"check_skills","agentperkdeck_playermanager_checkskills",function(self)
	if self:has_category_upgrade("player","agent_acrobatics_base") then 
		local upgrade_data = self:upgrade_value("player","agent_acrobatics_base")
		local default_duration = upgrade_data.duration
		local movement_grace_period = upgrade_data.movement_grace_period
		
		self._acrobatics_cached_is_moving_t = -movement_grace_period
		self._acrobatics_used = 0 --unique id for each acrobatics use and cooldown
		self._agent_acrobatics_timers = {}
		self._agent_acrobatics_cooldowns = {}
		Hooks:PostHook(self,"update","agent_acrobatics_update_handler",
			function(pm,t,dt)
				local any_expired
				local player = pm:local_player()
				if alive(player) then 
					if player:movement():current_state()._moving then 
						self._acrobatics_cached_is_moving_t = t
					end
				end
				
				local max_duration = 0
				local total_dodge_stacks = 0
				local total_crit_stacks = 0
				for i = #pm._agent_acrobatics_timers,1,-1 do 
					local charge_data = pm._agent_acrobatics_timers[i]
					charge_data.duration = charge_data.duration - dt
					if charge_data.duration < 0 then 
						any_expired = true
						table.remove(pm._agent_acrobatics_timers,i)
					else
						total_dodge_stacks = total_dodge_stacks + charge_data.dodge_stacks
						total_crit_stacks = total_crit_stacks + charge_data.crit_stacks
						max_duration = math_max(charge_data.duration,max_duration)
					end
				end
				if any_expired then 
					if #pm._agent_acrobatics_timers == 0 then 
						--this clears all crit and dodge stacks so it is not necessary to set either property here
						pm:on_all_acrobatics_charges_expired()
					else
						if pm:has_category_upgrade("player","agent_acrobatics_dodge_increase_on_crit") then
							pm:set_property("agent_acrobatics_dodge_stacks",total_dodge_stacks)
						end
						if pm:has_category_upgrade("player","agent_acrobatics_crit_chance_bonus") then
							pm:set_property("agent_acrobatics_crit_stacks",total_crit_stacks)
						end
					end
				end
				
				if max_duration ~= 0 then
					managers.hud:set_teammate_ability_radial(HUDManager.PLAYER_PANEL, {current=max_duration,total=math_max(max_duration,default_duration)})
				end
				
			end
		)
		
		self:set_property("agent_acrobatics_charges",self:upgrade_value("player","agent_acrobatics_charges_count",0)) --set starting charges
		
		local kill_extension = upgrade_data.kill_extension
		
		self:register_message(Message.OnEnemyKilled,"agent_acrobatics_on_enemy_killed",
			function(equipped_unit,variant,killed_unit)
				if self:is_acrobatics_active() then 
					--extend all timers by 0.5 seconds
					self:extend_acrobatics_timers(kill_extension)
				end
			end
		)
	else

		Hooks:RemovePostHook("agent_acrobatics_update_handler")
		
		self:remove_property("agent_acrobatics_charges")
		self:unregister_message(Message.OnEnemyKilled,"agent_acrobatics_extend_duration_on_kill")
	end
	
	if self:has_category_upgrade("player","agent_acrobatics_crit_chance_bonus") then 
		self:set_property("agent_acrobatics_crit_stacks",0)
		local upgrade_data = self:upgrade_value("player","agent_acrobatics_crit_chance_bonus")
		local max_stacks = upgrade_data.max_stacks
		
		self:register_message(Message.OnEnemyKilled,"agent_acrobatics_crit_chance_on_kill",
			function(equipped_unit,variant,killed_unit)
				if self:is_acrobatics_active() and self._agent_acrobatics_timers then
					local total_crit_stacks = 0
					for _,charge_data in pairs(self._agent_acrobatics_timers) do 
						local crit_stacks = charge_data.crit_stacks
						crit_stacks = math_min(max_stacks,crit_stacks + 1)
						total_crit_stacks = total_crit_stacks + crit_stacks
						charge_data.crit_stacks = crit_stacks
					end
					self:set_property("agent_acrobatics_crit_stacks",total_crit_stacks)
				end
			end
		)
	else
		self:remove_property("agent_acrobatics_crit_stacks")
		self:unregister_message(Message.OnEnemyKilled,"agent_acrobatics_crit_chance_on_kill")
	end
	
	if self:has_category_upgrade("player","agent_acrobatics_dodge_increase_on_crit") then 
		local upgrade_data = self:upgrade_value("player","agent_acrobatics_dodge_increase_on_crit")
		
		self:register_message(Message.OnEnemyShot,"agent_acrobatics_dodge_chance_on_crit",
			function(unit,attack_data) 
				if attack_data.critical_hit then
					if alive(attack_data.attacker_unit) and (attack_data.attacker_unit == self:local_player()) then 
						if self:is_acrobatics_active() and self._agent_acrobatics_timers then
						
							local max_stacks = upgrade_data.max_stacks
							local total_dodge_stacks = 0
							for _,charge_data in pairs(self._agent_acrobatics_timers) do 
								local dodge_stacks = charge_data.dodge_stacks
								dodge_stacks = math_min(max_stacks,dodge_stacks + 1)
								charge_data.dodge_stacks = dodge_stacks
								
								total_dodge_stacks = total_dodge_stacks + dodge_stacks
							end
							
							self:set_property("agent_acrobatics_dodge_stacks",total_dodge_stacks)
						end
					end
				end
			end
		)
	else
		self:unregister_message(Message.OnEnemyKilled,"agent_acrobatics_dodge_chance_on_crit")
	end
	
	if self:has_category_upgrade("player","agent_acrobatics_kill_cooldown_drain") then 
		local drain = self:upgrade_value("player","agent_acrobatics_kill_cooldown_drain")
		self:register_message(Message.OnEnemyKilled,"agent_acrobatics_cooldown_drain_on_kill",
			function(equipped_unit,variant,killed_unit)
				if not self:is_acrobatics_active(true) then 
					self:drain_acrobatics_cooldown(drain)
				end
			end
		)
	else
		self:unregister_message(Message.OnEnemyKilled,"agent_acrobatics_cooldown_drain_on_kill")
	end
	
	if self:has_category_upgrade("player","agent_acrobatics_heal_on_dodge") then 
		local heal_amount = self:upgrade_value("player","agent_acrobatics_heal_on_dodge")
		self:register_message(Message.OnPlayerDodge,"agent_acrobatics_proc_heal_on_dodge",
			function()
				if self:is_acrobatics_active() then 
					local player = self:local_player()
					if alive(player) then 
						player:character_damage():change_health(heal_amount * self:get_active_acrobatics_charges())
					end
				end
			end
		)
	else
		self:unregister_message(Message.OnPlayerDodge,"agent_acrobatics_proc_heal_on_dodge")
	end
	
	if self:has_category_upgrade("player","agent_acrobatics_dodge_inspire_basic_aoe") then 
		local dodge_threshold = self:upgrade_value("player","agent_acrobatics_dodge_inspire_basic_aoe")
		self:set_property("acrobatics_dodge_count",0)
		self:register_message(Message.OnPlayerDodge,"agent_acrobatics_increment_dodge_inspire_count",
			function()
				if self:is_acrobatics_active() then
					local dodge_count = self:get_property("acrobatics_dodge_count",0)
					dodge_count = dodge_count + 1 -- self:get_active_acrobatics_charges() to double activation threshold progress
					dodge_count = dodge_count % dodge_threshold
					if dodge_count == 0 then 

						--do inspire basic teamwide
						for _,data in pairs(managers.criminals:characters()) do 
							local unit = data.unit
							if alive(unit) and data.taken then
								unit:network():send_to_unit({
									"long_dis_interaction",
									unit,
									1, -- indicates basic
									self:local_player(),
									false
								})
--							elseif data.data.ai then 
								--inspire basic does not affect bots
							end
						end
					end
					self:set_property("acrobatics_dodge_count",dodge_count)
				end
			end
		)
	else
		self:set_property("acrobatics_dodge_count",0)
		self:unregister_message(Message.OnPlayerDodge,"agent_acrobatics_increment_dodge_inspire_count")
	end
end)

function PlayerManager:attempt_ability_acrobatics()
	if self:has_category_upgrade("player","agent_acrobatics_base") then
		local available_charges = self:get_property("agent_acrobatics_charges",0)
		if available_charges > 0 then
			local player = self:local_player()
			if alive(player) then --and not player:character_damage():is_downed() then 

				managers.hud:set_ability_icon(HUDManager.PLAYER_PANEL,"pd2_escape")
				
				local upgrade_data = self:upgrade_value("player","agent_acrobatics_base")
				local recharge_time = upgrade_data.cooldown
				local duration = upgrade_data.duration
				local max_charges = self:upgrade_value("player","agent_acrobatics_charges_count",0)
				
				self:add_to_property("agent_acrobatics_charges",-1)
				local used = self._acrobatics_used + 1
				self._acrobatics_used = used
				local timer_id = "agent_acrobatics_recharge_timer_" .. (used)
				
				self:start_timer(timer_id,recharge_time,
					function()
						self:add_to_property("agent_acrobatics_charges",1)
						
						if self:get_property("agent_acrobatics_charges") == max_charges then 
							self:local_player():sound():play("perkdeck_cooldown_over")
						end
					end
				)
				player:sound():play("perkdeck_activate")
				
				
				local charge_data = {
					aoe_stacks = 0, --number of times dodged from athletics, in order to proc inspire basic teamwide
					dodge_stacks = 0, --number of times crits scored from gymnastics, in order to grant bonus dodge chance
					crit_stacks = 0, --number of enemies killed from acrobatics, in order to grant bonus crit chance
					duration = duration,
					id = timer_id
				}
				
				table.insert(self._agent_acrobatics_cooldowns,timer_id)
				table.insert(self._agent_acrobatics_timers,charge_data)
			end
		end
	end
end

function PlayerManager:get_active_acrobatics_charges()
	return self._agent_acrobatics_timers and #self._agent_acrobatics_timers or 0
end

function PlayerManager:is_acrobatics_active(ignore_moving) --check acrobatics charge activated and all conditions valid for buffs to apply
	local player = self:local_player()
	local upgrade_data = self:upgrade_value("player","agent_acrobatics_base",{movement_grace_period = 0})
	local movement_grace_period = upgrade_data.movement_grace_period
	if alive(player) then 
		if ignore_moving or (self._acrobatics_cached_is_moving_t and (self._acrobatics_cached_is_moving_t < TimerManager:game():time() + movement_grace_period)) then
			if self._agent_acrobatics_timers then 
				return #self._agent_acrobatics_timers > 0
			end
		end
	end
	return false
end

function PlayerManager:drain_acrobatics_cooldown(drain)
--	local t = TimerManager:game():time()
	for i=#self._agent_acrobatics_cooldowns,1,-1 do
		local cooldown_id = self._agent_acrobatics_cooldowns[i]
		local timer_data = self._timers[cooldown_id]
		if timer_data then
			local expire_t = timer_data.t
			expire_t = expire_t - drain
			timer_data.t = expire_t
		else
			table.remove(self._agent_acrobatics_cooldowns,i)
		end
	end	
	--[[
	for _,charge_data in pairs(self._agent_acrobatics_timers) do 
		local timer_id = charge_data.id
		if self:get_timer(timer_id) then 
			local expire_t = self._timers[timer_id].t
			expire_t = expire_t - drain
			-- [ [ check timers immediately instead of waiting for next frame
			if t < expire_t then 
				self._timers[key].func(key,expire_t)
				self._timers[key] = nil
			else
				self._timers[timer_id].t = expire_t
			end
			-- ] ]
			self._timers[timer_id].t = expire_t
		end
	end
	--]]
end

function PlayerManager:extend_acrobatics_timers(extension)
	local upgrade_data = self:upgrade_value("player","agent_acrobatics_base",{duration=0})
	local max_duration = upgrade_data.duration
	for i,data in pairs(self._agent_acrobatics_timers) do 
		data.duration = math_min(data.duration + extension,max_duration)
	end
end

function PlayerManager:on_all_acrobatics_charges_expired() --when all stacks of acrobatics have expired
	self:set_property("agent_acrobatics_crit_stacks",0)
	self:set_property("agent_acrobatics_dodge_stacks",0)
	managers.hud:set_teammate_ability_radial(HUDManager.PLAYER_PANEL, {current=0,total=1})
end

local orig_critical_hit_chance = PlayerManager.critical_hit_chance 
function PlayerManager:critical_hit_chance(detection_risk,...)
	local bonus = 0
	if self:has_category_upgrade("player","agent_acrobatics_crit_chance_bonus") and self:is_acrobatics_active() then
		local upgrade_data = self:upgrade_value("player","agent_acrobatics_crit_chance_bonus")
		local stacks = self:get_property("agent_acrobatics_crit_stacks",0)
		local critical_chance_per_stack = upgrade_data.critical_hit_chance_bonus
		local crit_bonus = stacks * critical_chance_per_stack
		bonus = bonus + crit_bonus
	end
	return orig_critical_hit_chance(self,detection_risk,...) + bonus
end

local orig_skill_dodge_chance = PlayerManager.skill_dodge_chance 
function PlayerManager:skill_dodge_chance(running, crouching, on_zipline, override_armor, detection_risk,...)
	local bonus = 0
	if self:is_acrobatics_active() then
		
		if self:has_category_upgrade("player","agent_acrobatics_dodge_chance_bonus") then
			local dodge_chance_per_charge = self:upgrade_value("player","agent_acrobatics_dodge_chance_bonus",0)
			local dodge_bonus = dodge_chance_per_charge * self:get_active_acrobatics_charges()
			bonus = bonus + dodge_bonus
		end
		
		if self:has_category_upgrade("player","agent_acrobatics_dodge_increase_on_crit") and self._agent_acrobatics_timers then 
			local upgrade_value = self:upgrade_value("player","agent_acrobatics_dodge_increase_on_crit")
			local dodge_chance_per_stack = upgrade_value.dodge_chance_bonus
			
			local stacks = self:get_property("agent_acrobatics_dodge_stacks",0)
			local dodge_bonus = dodge_chance_per_stack * stacks
			bonus = bonus + dodge_bonus
		end
	end
	
	return orig_skill_dodge_chance(self,running, crouching, on_zipline, override_armor, detection_risk,...) + bonus
end
